import csv
import io
import json
import logging
import os

from collections import defaultdict

from ..utils.exceptions import OasisException
from ..utils.log import oasis_log
from .files import GENERAL_SETTINGS_FILE, GUL_SUMMARIES_FILE, IL_SUMMARIES_FILE, MODEL_SETTINGS_FILE


def _get_summaries(summary_file):
    """
    Get a list representation of a summary file.
    """
    summaries_dict = defaultdict(lambda: {'leccalc': {}})

    with io.open(summary_file, 'r', encoding='utf-8') as csvfile:
        reader = csv.reader(csvfile)
        for row in reader:
            id = int(row[0])

            if row[1].startswith('leccalc'):
                summaries_dict[id]['leccalc'][row[1]] = row[2].lower() == 'true'
            else:
                summaries_dict[id][row[1]] = row[2].lower() == 'true'

    summaries = list()
    for id in sorted(summaries_dict):
        summaries_dict[id]['id'] = id
        summaries.append(summaries_dict[id])

    return summaries


@oasis_log
def create_analysis_settings_json(directory):
    """
    Generate an analysis settings JSON from a set of
    CSV files in a specified directory.
    Args:
        ``directory`` (string): the directory containing the CSV files.
    Returns:
        The analysis settings JSON.
    """
    if not os.path.exists(directory):
        error_message = "Directory does not exist: {}".format(directory)
        logging.getLogger().error(error_message)
        raise OasisException(error_message)

    general_settings_file = os.path.join(directory, GENERAL_SETTINGS_FILE)
    model_settings_file = os.path.join(directory, MODEL_SETTINGS_FILE)
    gul_summaries_file = os.path.join(directory, GUL_SUMMARIES_FILE)
    il_summaries_file = os.path.join(directory, IL_SUMMARIES_FILE)

    for file in [general_settings_file, model_settings_file, gul_summaries_file, il_summaries_file]:
        if not os.path.exists(file):
            error_message = "File does not exist: {}".format(directory)
            logging.getLogger().error(error_message)
            raise OasisException(error_message)

    general_settings = dict()
    with io.open(general_settings_file, 'r', encoding='utf-8') as csvfile:
        reader = csv.reader(csvfile)
        for row in reader:
            general_settings[row[0]] = eval("{}('{}')".format(row[2], row[1]))

    model_settings = dict()
    with io.open(model_settings_file, 'r', encoding='utf-8') as csvfile:
        reader = csv.reader(csvfile)
        for row in reader:
            model_settings[row[0]] = eval("{}('{}')".format(row[2], row[1]))

    gul_summaries = _get_summaries(gul_summaries_file)
    il_summaries = _get_summaries(il_summaries_file)

    analysis_settings = general_settings
    analysis_settings['model_settings'] = model_settings
    analysis_settings['gul_summaries'] = gul_summaries
    analysis_settings['il_summaries'] = il_summaries
    output_json = json.dumps(analysis_settings)
    logging.getLogger().info("Analysis settings json: {}".format(output_json))

    return output_json



import argparse
import logging

from .getmodel import manager, logger


parser = argparse.ArgumentParser()
parser.add_argument('-i', '--file-in', help='names of the input file_path')
parser.add_argument('-o', '--file-out', help='names of the output file_path')
parser.add_argument('-r', '--run-dir', help='path to the run directory', default='.')
parser.add_argument('--ignore-file-type', nargs='*', help='the type of file to be loaded', default=set())
parser.add_argument('-v', '--logging-level', help='logging level (debug:10, info:20, warning:30, error:40, critical:50)',
                    default=30, type=int)


def main() -> None:
    """
    Is the entry point for the modelpy command which loads data and constructs a model.

    Returns: None
    """
    kwargs = vars(parser.parse_args())

    # add handler to fm logger
    ch = logging.StreamHandler()
    formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    ch.setFormatter(formatter)
    logger.addHandler(ch)
    logging_level = kwargs.pop('logging_level')
    logger.setLevel(logging_level)

    manager.run(**kwargs)


if __name__ == "__main__":
    main()



__all__ = [
    'get_config_profile',
    'get_default_accounts_profile',
    'get_default_deterministic_analysis_settings',
    'get_default_exposure_profile',
    'get_default_fm_profile_field_values',
    'get_default_step_policies_profile',
    'get_default_fm_aggregation_profile',
    'get_default_unified_profile',
    'get_loc_dtypes',
    'get_acc_dtypes',
    'get_scope_dtypes',
    'get_info_dtypes',
    'get_oed_default_values',
    'assign_defaults_to_il_inputs',
    'store_exposure_fp',
    'find_exposure_fp',
    'GROUP_ID_COLS',
    'CORRELATION_GROUP_ID',
    'API_EXAMPLE_AUTH',
    'DEFAULT_RTREE_INDEX_PROPS',
    'KTOOLS_ALLOC_GUL_MAX',
    'KTOOLS_ALLOC_FM_MAX',
    'KTOOLS_FIFO_RELATIVE',
    'KTOOLS_DEBUG',
    'KTOOLS_DISABLE_ERR_GUARD',
    'KTOOLS_NUM_PROCESSES',
    'KTOOLS_GUL_LEGACY_STREAM',
    'OASIS_FILES_PREFIXES',
    'SUMMARY_MAPPING',
    'SUMMARY_OUTPUT',
    'SOURCE_IDX',
    'STATIC_DATA_FP',
    'WRITE_CHUNKSIZE'
]

import os
import io
import glob
import json

from collections import OrderedDict
from itertools import chain

from .fm import SUPPORTED_FM_LEVELS
from .exceptions import OasisException

try:
    from json import JSONDecodeError
except ImportError:
    from builtins import ValueError as JSONDecodeError

SOURCE_FILENAMES = OrderedDict({
    'loc': 'location.csv',
    'acc': 'account.csv',
    'info': 'ri_info.csv',
    'scope': 'ri_scope.csv',
    'complex_lookup': 'analysis_settings.json',
    'oed_location_csv': 'location.csv',
    'oed_accounts_csv': 'account.csv',
    'oed_info_csv': 'ri_info.csv',
    'oed_scope_csv': 'ri_scope.csv',
    'lookup_config_json': 'lookup.json',
    'profile_loc_json': 'profile_location.json',
    'keys_data_csv': 'keys.csv',
    'model_version_csv': 'model_version.csv',
    'lookup_complex_config_json': 'lookup_complex.json',
    'profile_acc_json': 'profile_account.json',
    'profile_fm_agg_json': 'profile_fm_agg.json',
})

API_EXAMPLE_AUTH = OrderedDict({
    'user': 'admin',
    'pass': 'password',
})

DEFAULT_RTREE_INDEX_PROPS = {
    'buffering_capacity': 10,
    'custom_storage_callbacks': None,
    'custom_storage_callbacks_size': 0,
    'dat_extension': 'dat',
    'dimension': 2,
    'filename': '',
    'fill_factor': 0.7,
    'idx_extension': 'idx',
    'index_capacity': 100,
    'index_id': None,
    'leaf_capacity': 100,
    'near_minimum_overlap_factor': 32,
    'overwrite': True,
    'pagesize': 4096,
    'point_pool_capacity': 500,
    'region_pool_capacity': 1000,
    'reinsert_factor': 0.3,
    'split_distribution_factor': 0.4,
    'storage': 0,
    'tight_mbr': True,
    'tpr_horizon': 20.0,
    'type': 0,
    'variant': 2,
    'writethrough': False
}


# Store index from merged source files (for later slice & dice)
SOURCE_IDX = OrderedDict({
    'loc': 'loc_idx',
    'acc': 'acc_idx',
    'info': 'info_idx',
    'scope': 'scope_idx'
})

SUMMARY_MAPPING = OrderedDict({
    'gul_map_fn': 'gul_summary_map.csv',
    'fm_map_fn': 'fm_summary_map.csv'
})

SUMMARY_OUTPUT = OrderedDict({
    'gul': 'gulsummaryxref.csv',
    'il': 'fmsummaryxref.csv'
})

# Path for storing static data/metadata files used in the package
STATIC_DATA_FP = os.path.join(os.path.abspath(os.path.join(os.path.dirname(__file__), os.pardir)), '_data')

# Default profiles that describe the financial terms in the OED acc. and loc.
# (exposure) files, as well as how aggregation of FM input items is performed
# in the different OED FM levels


def store_exposure_fp(fp, exposure_type):
    """
    Preserve original exposure file extention if its in a pandas supported
    compressed format

    compression : {‘infer’, ‘gzip’, ‘bz2’, ‘zip’, ‘xz’, None}, default ‘infer’
                  For on-the-fly decompression of on-disk data. If ‘infer’ and
                  filepath_or_buffer is path-like, then detect compression from
                  the following extensions: ‘.gz’, ‘.bz2’, ‘.zip’, or ‘.xz’
                  (otherwise no decompression).

                  If using ‘zip’, the ZIP file must contain only one data file
                  to be read in. Set to None for no decompression.

                New in version 0.18.1: support for ‘zip’ and ‘xz’ compression.
    """
    compressed_ext = ('.gz', '.bz2', '.zip', '.xz', '.parquet')
    filename = SOURCE_FILENAMES[exposure_type]
    if fp.endswith(compressed_ext):
        return '.'.join([filename, fp.rsplit('.')[-1]])
    else:
        return filename


def find_exposure_fp(input_dir, exposure_type, required=True):
    """
    Find an OED exposure file stored in the oasis inputs dir
    while preserving the compressed ext
    """
    fp = glob.glob(os.path.join(input_dir, SOURCE_FILENAMES[exposure_type].rsplit(".", 1)[0] + '*'))
    if required or fp:
        return fp.pop()


def get_default_json(src_fp):
    """
    Loads JSON from file.

    :param src_fp: Source JSON file path
    :type src_fp: str

    :return: dict
    :rtype: dict
    """
    try:
        with io.open(src_fp, 'r', encoding='utf-8') as f:
            return json.load(f)
    except (IOError, JSONDecodeError, OSError, TypeError):
        raise OasisException('Error trying to load JSON from {}'.format(src_fp))


def get_default_accounts_profile(path=False):
    fp = os.path.join(STATIC_DATA_FP, 'default_acc_profile.json')
    return get_default_json(src_fp=fp) if not path else fp


def get_default_exposure_profile(path=False):
    fp = os.path.join(STATIC_DATA_FP, 'default_loc_profile.json')
    return get_default_json(src_fp=fp) if not path else fp


def get_default_fm_profile_field_values(path=False):
    fp = os.path.join(STATIC_DATA_FP, 'default_fm_profile_field_values.json')
    return get_default_json(src_fp=fp) if not path else fp


def get_default_step_policies_profile(path=False):
    fp = os.path.join(STATIC_DATA_FP, 'default_step_policies_profile.json')
    return get_default_json(src_fp=fp) if not path else fp


def get_config_profile(path=False):
    fp = os.path.join(STATIC_DATA_FP, 'config_compatibility_profile.json')
    return get_default_json(src_fp=fp) if not path else fp


def get_default_unified_profile(path=False):
    fp = os.path.join(STATIC_DATA_FP, 'default_unified_profile.json')
    return get_default_json(src_fp=fp) if not path else fp


def get_default_fm_aggregation_profile(path=False):
    fp = os.path.join(STATIC_DATA_FP, 'default_fm_agg_profile.json')
    return {int(k): v for k, v in get_default_json(src_fp=fp).items()} if not path else fp


def get_loc_dtypes():
    fp = os.path.join(STATIC_DATA_FP, 'loc_dtypes.json')
    return get_default_json(src_fp=fp)


def get_acc_dtypes():
    fp = os.path.join(STATIC_DATA_FP, 'acc_dtypes.json')
    return get_default_json(src_fp=fp)


def get_scope_dtypes():
    fp = os.path.join(STATIC_DATA_FP, 'scope_dtypes.json')
    return get_default_json(src_fp=fp)


def get_info_dtypes():
    fp = os.path.join(STATIC_DATA_FP, 'info_dtypes.json')
    return get_default_json(src_fp=fp)


def get_oed_default_values(terms):
    """
    Get defaults value of OED terms.

    :param terms: list of OED terms
    :type terms: list

    :return: default values for OED terms
    :rtype: dict
    """
    loc_defaults = {k.lower(): v['default_value'] for k, v in get_loc_dtypes().items() if k.lower() in terms}
    acc_defaults = {k.lower(): v['default_value'] for k, v in get_acc_dtypes().items() if k.lower() in terms}
    defaults = dict(
        chain.from_iterable(d.items() for d in (loc_defaults, acc_defaults))
    )

    return defaults


def assign_defaults_to_il_inputs(df):
    """
    Assign default values to IL inputs.

    :param df: IL input items dataframe
    :type df: pandas.DataFrame

    :return: IL input items dataframe
    :rtype: pandas.DataFrame
    """
    # Get default values for IL inputs
    default_fm_profile_field_values = get_default_fm_profile_field_values()

    for level in default_fm_profile_field_values.keys():
        level_id = SUPPORTED_FM_LEVELS[level]['id']
        for k, v in default_fm_profile_field_values[level].items():
            # Evaluate condition for assigning default values if present
            if v.get('condition'):
                df.loc[df.level_id == level_id, k] = df.loc[
                    df.level_id == level_id, k
                ].where(eval(
                    'df.loc[df.level_id == level_id, k]' + v['condition']),
                    v['default_value']
                )
            else:
                df.loc[df.level_id == level_id, k] = v['default_value']

    return df


WRITE_CHUNKSIZE = 2 * (10 ** 5)

GROUP_ID_COLS = ['loc_id']

CORRELATION_GROUP_ID = ['CorrelationGroup']

# Default name prefixes of the Oasis input files (GUL + IL)
OASIS_FILES_PREFIXES = OrderedDict({
    'gul': {
        'complex_items': 'complex_items',
        'items': 'items',
        'coverages': 'coverages',
    },
    'il': {
        'fm_policytc': 'fm_policytc',
        'fm_profile': 'fm_profile',
        'fm_programme': 'fm_programme',
        'fm_xref': 'fm_xref',
    }
})


# Default analysis settings for deterministic loss generation
def get_default_deterministic_analysis_settings(path=False):
    fp = os.path.join(STATIC_DATA_FP, 'analysis_settings.json')
    return get_default_json(src_fp=fp) if not path else fp


# Defaults for Ktools runtime parameters
KTOOLS_NUM_PROCESSES = -1
KTOOLS_FIFO_RELATIVE = False
KTOOLS_DISABLE_ERR_GUARD = False
KTOOLS_GUL_LEGACY_STREAM = False
# ktools gul alloc rules:
# 2 = total loss is maximum subperil loss
# 1 = default with back allocation
# 0 = default without back allocation
KTOOLS_ALLOC_GUL_MAX = 2
KTOOLS_ALLOC_FM_MAX = 3
KTOOLS_ALLOC_GUL_DEFAULT = 0
KTOOLS_ALLOC_IL_DEFAULT = 2
KTOOLS_ALLOC_RI_DEFAULT = 3
KTOOLS_TIV_SAMPLE = -2
KTOOLS_MEAN_SAMPLE_IDX = -1
KTOOLS_STD_DEV_SAMPLE_IDX = -2
KTOOLS_TIV_SAMPLE_IDX = -3
KTOOL_N_GUL_PER_LB = 0
KTOOL_N_FM_PER_LB = 0

# Values for event shuffle rules
EVE_NO_SHUFFLE = 0
EVE_ROUND_ROBIN = 1
EVE_FISHER_YATES = 2
EVE_STD_SHUFFLE = 3
EVE_DEFAULT_SHUFFLE = EVE_ROUND_ROBIN

KTOOLS_DEBUG = False



__all__ = [
    'PlatformBase',
    'PlatformList',
    'PlatformGet',
    'PlatformRun',
    'PlatformDelete',
]


import getpass
import io
import os
import json
import sys

from tabulate import tabulate
from mimetypes import guess_extension

from requests.exceptions import HTTPError, ConnectionError

from ...platform.client import APIClient
from ...utils.exceptions import OasisException
from ...utils.defaults import API_EXAMPLE_AUTH

from ..base import ComputationStep


class PlatformBase(ComputationStep):
    """
    Base platform class to handle opening a client connection
    """
    step_params = [
        {'name': 'server_login_json', 'is_path': True, 'pre_exist': False, 'help': 'Source location CSV file path'},
        {'name': 'server_url', 'default': 'http://localhost:8000', 'help': 'URL to Oasis Platform server, default is localhost'},
        {'name': 'server_version', 'default': 'V1'},
    ]

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.server = self.open_connection()

    def load_credentials(self, login_arg):
        """
        Load credentials from JSON file or Prompt for username / password

        Options:
            1.'--server-login ./APIcredentials.json'
            2. Load credentials from default config file '-C oasislmf.json'
        """
        if isinstance(login_arg, str):
            with io.open(login_arg, encoding='utf-8') as f:
                return json.load(f)
        elif isinstance(login_arg, dict):
            if {'password', 'username'} <= {k for k in login_arg.keys()}:
                return login_arg
        else:
            self.logger.info('API Login:')

        try:
            api_login = {}
            api_login['username'] = input('Username: ')
            api_login['password'] = getpass.getpass('Password: ')
            return api_login
        except KeyboardInterrupt as e:
            self.logger.error('\nFailed to get API login details:')
            self.logger.error(e)
            sys.exit(1)

    def open_connection(self):
        try:
            # If no password given try the reference example
            return APIClient(
                api_url=self.server_url,
                api_ver=self.server_version,
                username=API_EXAMPLE_AUTH['user'],
                password=API_EXAMPLE_AUTH['pass'],
            )
        except OasisException as e:
            if isinstance(e.original_exception, HTTPError):
                # Prompt for password and try to re-autehnticate
                self.logger.info("-- Authentication Required --")
                credentials = self.load_credentials(self.server_login_json)
                self.logger.info(f'Connecting to - {self.server_url}')
                return APIClient(
                    api_url=self.server_url,
                    api_ver=self.server_version,
                    username=credentials['username'],
                    password=credentials['password'],
                )
            elif isinstance(e.original_exception, ConnectionError):
                self.logger.info('API Connection error to "{}"'.format(self.server_url))
                self.logger.debug(e)
                sys.exit(1)
            else:
                self.logger.error('Unhandled error:')
                self.logger.debug(e)
                sys.exit(1)

    def tabulate_json(self, json_data, items):
        table_data = dict()

        for i in items:
            table_data[i] = list()

        for m in json_data:
            for k in table_data:
                # will have link+data if dict returned
                if isinstance(m[k], dict):
                    table_data[k].append('Yes')

                # If none then no data
                elif m[k] is None:
                    table_data[k].append('-')

                # If URL then something linked to field
                elif isinstance(m[k], str):
                    if any(v in m[k] for v in ['http://', 'https://']):
                        table_data[k].append('Linked')
                    else:
                        table_data[k].append(m[k])

                # Fallback - add value as string
                else:
                    table_data[k].append(str(m[k]))
        return table_data

    def print_endpoint(self, attr, items):
        endpoint_obj = getattr(self.server, attr)

        self.logger.info(f'\nAvailable {attr}:')
        data = self.tabulate_json(endpoint_obj.get().json(), items)
        self.logger.info(tabulate(data, headers=items, tablefmt='psql'))
        return data

    def run(self):
        """ Add the logic for platform API interaction here
        """
        raise NotImplementedError('Methode run must be implemented, this class handles base server connection')


class PlatformList(PlatformBase):
    """ Return status and details from an Oasis Platform API server
    """
    step_params = PlatformBase.step_params + [
        {'name': 'models', 'flag': '-m', 'type': int, 'nargs': '+', 'help': 'List of model ids to print in detail'},
        {'name': 'portfolios', 'flag': '-p', 'type': int, 'nargs': '+', 'help': 'List of portfolio ids to print in detail'},
        {'name': 'analyses', 'flag': '-a', 'type': int, 'nargs': '+', 'help': 'List of analyses ids to print in detail'},
    ]

    def run(self):
        # Default to printing summary of API status
        if not any([self.models, self.portfolios, self.analyses]):
            self.print_endpoint('models', ['id', 'supplier_id', 'model_id', 'version_id'])
            self.print_endpoint('portfolios', ['id', 'name', 'location_file', 'accounts_file', 'reinsurance_info_file', 'reinsurance_scope_file'])
            self.print_endpoint('analyses', ['id', 'name', 'model', 'portfolio', 'status', 'input_file', 'output_file', 'run_log_file'])

        if self.models:
            for Id in self.models:
                msg = f'Model (id={Id}): \n'
                try:
                    rsp = self.server.models.get(Id)
                    self.logger.info(msg + json.dumps(rsp.json(), indent=4, sort_keys=True))
                except HTTPError as e:
                    self.logger.info(msg + e.response.text)

        if self.portfolios:
            for Id in self.portfolios:
                msg = f'Portfolio (id={Id}): \n'
                try:
                    rsp = self.server.portfolios.get(Id)
                    self.logger.info(msg + json.dumps(rsp.json(), indent=4, sort_keys=True))
                except HTTPError as e:
                    self.logger.info(msg + e.response.text)

        if self.analyses:
            for Id in self.analyses:
                msg = f'Analysis (id={Id}): \n'
                try:
                    rsp = self.server.analyses.get(Id)
                    self.logger.info(msg + json.dumps(rsp.json(), indent=4, sort_keys=True))
                except HTTPError as e:
                    self.logger.info(msg + e.response.text)


class PlatformRun(PlatformBase):
    """ End to End - run model via the Oasis Platoform API
    """
    step_params = PlatformBase.step_params + [
        {'name': 'model_id',     'type': int, 'help': 'API `id` of a model to run an analysis with'},
        {'name': 'portfolio_id', 'type': int, 'help': 'API `id` of a portfolio to run an analysis with'},
        {'name': 'analysis_id',  'type': int, 'help': 'API `id` of an analysis to run'},

        {'name': 'output_dir',             'flag': '-o', 'is_path': True, 'pre_exist': True, 'help': 'Output data directory for results data (absolute or relative file path)', 'default': './'},
        {'name': 'analysis_settings_json', 'flag': '-a', 'is_path': True, 'pre_exist': True, 'help': 'Analysis settings JSON file path'},
        {'name': 'oed_location_csv',       'flag': '-x', 'is_path': True, 'pre_exist': True, 'help': 'Source location CSV file path'},
        {'name': 'oed_accounts_csv',       'flag': '-y', 'is_path': True, 'pre_exist': True, 'help': 'Source accounts CSV file path'},
        {'name': 'oed_info_csv',           'flag': '-i', 'is_path': True, 'pre_exist': True, 'help': 'Reinsurance info. CSV file path'},
        {'name': 'oed_scope_csv',          'flag': '-s', 'is_path': True, 'pre_exist': True, 'help': 'Reinsurance scope CSV file path'},
    ]

    def select_id(self, msg, valid_ids):
        while True:
            try:
                value = int(input(f'Select {msg} ID: '))
            except ValueError:
                self.logger.info('Invalid Response: {}'.format(value))
                continue
            except KeyboardInterrupt:
                exit(1)

            if (value < 0) or (str(value) not in valid_ids):
                self.logger.info('Not a valid id from: {}'.format(valid_ids))
                continue
            else:
                break
        return value

    def run(self):
        # if given an analysis id, select that with new settings file (If given)
        if self.analysis_id:
            try:
                status = self.server.analyses.status(self.analysis_id)
                if status in ['RUN_QUEUED', 'RUN_STARTED']:
                    self.server.cancel_analysis(self.analysis_id)
                elif status in ['INPUTS_GENERATION_QUEUED', 'INPUTS_GENERATION_STARTED']:
                    self.server.cancel_generate(self.analysis_id)
                analysis_id = self.analysis_id
            except HTTPError as e:
                self.logger.error('Error running analysis ({}) - {}'.format(
                    self.analysis_id, e))
                sys.exit(1)
        else:
            # If not given an analsis then we need to select a model + portfolio
            # when no model is selected prompt user for choice of installed models
            if not self.model_id:
                models = self.server.models.get().json()

                if len(models) > 1:
                    models_table = self.print_endpoint('models', ['id', 'supplier_id', 'model_id', 'version_id'])
                    model_id = self.select_id('models', models_table['id'])
                elif len(models) == 1:
                    model_id = models[0]['id']
                else:
                    raise OasisException(
                        'No models found in API: {}'.format(self.server_url)
                    )

            # Select or create a portfilo
            if self.portfolio_id:
                portfolios = self.server.portfolios.get().json()
                valid_portfolios = self.tabulate_json(portfolios, ['id'])
                portfolio_id = self.portfolio_id
                if str(self.portfolio_id) not in valid_portfolios['id']:
                    raise OasisException(f'Run Error: Invalid selected value "{self.portfolio_id}" for portfolio_id:')

            elif self.oed_location_csv:
                portfolio = self.server.upload_inputs(
                    portfolio_id=None,
                    location_fp=self.oed_location_csv,
                    accounts_fp=self.oed_accounts_csv,
                    ri_info_fp=self.oed_info_csv,
                    ri_scope_fp=self.oed_scope_csv,
                )
                portfolio_id = portfolio['id']
            else:
                raise OasisException('Error: Either select a "portfolio_id" or Input exposure files to run.')
            analysis = self.server.create_analysis(
                portfolio_id=portfolio_id,
                model_id=model_id,
                analysis_settings_fp=self.analysis_settings_json,
            )
            analysis_id = analysis['id']

        # Execure run
        self.server.run_generate(analysis_id)
        self.server.run_analysis(analysis_id, self.analysis_settings_json)
        self.server.download_output(analysis_id, self.output_dir)


class PlatformDelete(PlatformBase):
    """ Delete either a 'model', 'portfolio' or an 'analysis' from the API's Database
    """
    step_params = PlatformBase.step_params + [
        {'name': 'models', 'flag': '-m', 'type': int, 'nargs': '+', 'help': 'List of model ids to Delete.'},
        {'name': 'portfolios', 'flag': '-p', 'type': int, 'nargs': '+', 'help': 'List of Portfolio ids to Delete'},
        {'name': 'analyses', 'flag': '-a', 'type': int, 'nargs': '+', 'help': 'List of Analyses ids to Detele'},
    ]

    def delete_list(self, attr, id_list):
        api_endpoint = getattr(self.server, attr)
        for Id in id_list:
            try:
                if api_endpoint.delete(Id).ok:
                    self.logger.info(f'Deleted {attr}_id={Id}')
            except HTTPError as e:
                self.logger.error('Delete error {}_id={} - {}'.format(attr, Id, e))
                continue

    def run(self):

        if not any([self.models, self.portfolios, self.analyses]):
            raise OasisException("""Select item(s) to delete, list of either:
                --models MODEL_ID [MODEL_ID ... n],
                --portfolios PORTFOLIOS_ID [PORTFOLIO_ID ... n]
                --analyses ANALYSES_ID [ANALYSES_ID ... n]
                """)

        if self.models:
            self.delete_list('models', self.models)
        if self.portfolios:
            self.delete_list('portfolios', self.portfolios)
        if self.analyses:
            self.delete_list('analyses', self.analyses)


class PlatformGet(PlatformBase):
    """ Download file(s) from the api
    """
    step_params = PlatformBase.step_params + [
        {'name': 'output_dir', 'flag': '-o', 'is_path': True, 'pre_exist': True, 'help': 'Output data directory for results data (absolute or relative file path)', 'default': './'},
        # Files for models object
        {'name': 'model_settings',     'type': int, 'nargs': '+', 'help': 'Model ids to download settings file.'},
        {'name': 'model_versions',     'type': int, 'nargs': '+', 'help': 'Model ids to download versions file'},
        # Files from portfolio
        {'name': 'portfolio_location_file',          'type': int, 'nargs': '+', 'help': 'Portfolio ids to download Location file'},
        {'name': 'portfolio_accounts_file',          'type': int, 'nargs': '+', 'help': 'Portfolio ids to download Accounts file'},
        {'name': 'portfolio_reinsurance_scope_file', 'type': int, 'nargs': '+', 'help': 'Portfolio ids to download RI scope file.'},
        {'name': 'portfolio_reinsurance_info_file',  'type': int, 'nargs': '+', 'help': 'Portfolio ids to download RI info file.'},
        # Files from an analyses
        {'name': 'analyses_settings_file',                   'type': int, 'nargs': '+', 'help': 'Analyses ids to download settings file'},
        {'name': 'analyses_run_traceback_file',              'type': int, 'nargs': '+', 'help': 'Analyses ids to download traceback logs'},
        {'name': 'analyses_run_log_file',                    'type': int, 'nargs': '+', 'help': 'Analyses ids to download Ktools run logs'},
        {'name': 'analyses_input_generation_traceback_file', 'type': int, 'nargs': '+', 'help': 'Analyses ids to download input_generation traceback logs'},
        {'name': 'analyses_input_file',                      'type': int, 'nargs': '+', 'help': 'Analyses ids to download Generated inputs tar'},
        {'name': 'analyses_output_file',                     'type': int, 'nargs': '+', 'help': 'Analyses ids to download Output losses tar'},
        {'name': 'analyses_summary_levels_file',             'type': int, 'nargs': '+', 'help': 'Analyses ids to download summary levels file'},
        {'name': 'analyses_lookup_validation_file',          'type': int, 'nargs': '+', 'help': 'Analyses ids to download exposure summary'},
        {'name': 'analyses_lookup_success_file',             'type': int, 'nargs': '+', 'help': 'Analyses ids to download successful lookups'},
        {'name': 'analyses_lookup_errors_file',              'type': int, 'nargs': '+', 'help': 'Analyses ids to download summary of failed lookups'},
    ]

    def extract_args(self, param_suffix):
        return {k.replace(param_suffix, ''): v for k, v in self.kwargs.items() if v and param_suffix in k}

    def download(self, collection, req_files, chuck_size=1024):
        collection_obj = getattr(self.server, collection)

        for File in req_files:
            resource = getattr(collection_obj, File)
            for ID in req_files[File]:
                try:
                    r = resource.get(ID)

                    ext = guess_extension(r.headers['content-type'].partition(';')[0].strip())
                    filename = os.path.join(self.output_dir, f'{ID}_{collection}_{File}{ext}')

                    with io.open(filename, 'wb') as f:
                        for chunk in r.iter_content(chunk_size=chuck_size):
                            f.write(chunk)
                    self.logger.info(f'Downloaded: {File} from {collection}_id={ID} "{filename}"')
                except HTTPError as e:
                    self.logger.error('Download failed: - {}'.format(e))

    def run(self):
        model_files = self.extract_args('model_')
        portfolio_files = self.extract_args('portfolio_')
        analyses_files = self.extract_args('analyses_')

        # Check that at least one option is given
        if not any([model_files, portfolio_files, analyses_files]):
            raise OasisException('Select file for download e.g. "--analyses_output <id_1> .. <id_n>"')

        if model_files:
            self.download('models', model_files)
        if portfolio_files:
            self.download('portfolios', portfolio_files)
        if analyses_files:
            self.download('analyses', analyses_files)



__all__ = [
    'ExposureCmd',
    'RunCmd'
]

import os

from argparse import RawDescriptionHelpFormatter

from .command import OasisBaseCommand, OasisComputationCommand


class RunCmd(OasisComputationCommand):
    """
    Generates deterministic losses using the installed ktools framework given
    direct Oasis files (GUL + optionally IL and RI input files).

    The command line arguments can be supplied in the configuration file
    (``oasislmf.json`` by default or specified with the ``--config`` flag).
    Run ``oasislmf config --help`` for more information.
    """
    formatter_class = RawDescriptionHelpFormatter
    computation_name = 'RunExposure'


class ExposureCmd(OasisBaseCommand):
    """
    Exposure subcommands::

        * generate - and optionally, validate - deterministic losses (GUL, IL or RIL)
    """
    sub_commands = {
        'run': RunCmd
    }



__all__ = [
    'RunModel'
]

import os
from tqdm import tqdm

from ..base import ComputationStep

from ..generate.files import GenerateFiles
from ..generate.losses import GenerateLosses
from ..hooks.pre_analysis import ExposurePreAnalysis

from ...utils.exceptions import OasisException
from ...utils.data import (
    get_analysis_settings,
    get_model_settings,
)

from ...utils.path import empty_dir
from ...utils.defaults import store_exposure_fp

class RunModel(ComputationStep):
    """
    Run models end to end.
    """

    # Override params
    step_params = [
        {'name': 'oasis_files_dir', 'flag': '-o', 'is_path': True, 'pre_exist': False, 'help': 'Path to the directory in which to generate the Oasis files'},
        {'name': 'exposure_pre_analysis_module', 'required': False, 'is_path': True, 'pre_exist': True, 'help': 'Exposure Pre-Analysis lookup module path'},
    ]
    # Add params from each sub command not in 'step_params'
    chained_commands = [
        GenerateLosses,
        GenerateFiles,
        ExposurePreAnalysis,
    ]


    def pre_analysis_kwargs(self):
        updated_inputs = {}
        input_dir = self.kwargs['oasis_files_dir']

        for input_name in ('oed_location_csv', 'oed_accounts_csv', 'oed_info_csv', 'oed_scope_csv'):
            if self.kwargs[input_name]:
                updated_inputs[input_name] = os.path.join(
                    input_dir,
                    store_exposure_fp(self.kwargs[input_name], input_name)
                )
        return {**self.kwargs, **updated_inputs}


    def run(self):

        # setup output dir
        if not self.model_run_dir:
            self.model_run_dir = GenerateLosses._get_output_dir(self)
        if os.path.exists(self.model_run_dir):
            empty_dir(self.model_run_dir)
        os.makedirs(os.path.join(self.model_run_dir, 'input'))

        self.kwargs['model_run_dir'] = self.model_run_dir
        self.kwargs['oasis_files_dir'] = os.path.join(self.model_run_dir, 'input')

        # Validate JSON files (Fail at entry point not after input generation)
        get_analysis_settings(self.analysis_settings_json)
        if self.model_settings_json:
            get_model_settings(self.model_settings_json)

        # Check input expsoure
        required_ri_paths = [self.oed_info_csv, self.oed_scope_csv]
        il = True if self.oed_accounts_csv else False
        ri = all(required_ri_paths) and il
        if any(required_ri_paths) and not ri:
            raise OasisException(
                'RI option indicated by provision of some RI related assets, but other assets are missing. '
                'To generate RI inputs you need to provide all of the assets required to generate direct '
                'Oasis files (GUL + FM input files) plus all of the following assets: '
                '    reinsurance info. file path, '
                '    reinsurance scope file path.'
            )

        # Run chain
        if self.exposure_pre_analysis_module:
            cmds = [(ExposurePreAnalysis, self.kwargs), (GenerateFiles, self.pre_analysis_kwargs()), (GenerateLosses, self.kwargs)]
        else:
            cmds = [(GenerateFiles, self.kwargs), (GenerateLosses, self.kwargs)]

        with tqdm(total=len(cmds)) as pbar:
            for cmd in cmds:
                cmd[0](**cmd[1]).run()
                pbar.update(1)

        self.logger.info('\nModel run completed successfully in {}'.format(self.model_run_dir))



__all__ = [
    "CreateModelRepo",
    "CreateComplexModelRepo",
]

from subprocess import (
    CalledProcessError,
    run
)

from ..base import ComputationStep
from ...utils.exceptions import OasisException

try:
    import cookiecutter
    cookiecutter_ver = cookiecutter.__version__
except ImportError:
    cookiecutter_ver = None


class CreateModelBase(ComputationStep):
    """
    Creates a local Git repository for a "simple model" (using the
    ``cookiecutter`` package) on a "simple model" repository template
    on GitHub)
    """

    step_params = [
        {'name': 'output_dir', 'flag': '-o', 'is_path': True, 'pre_exist': False, 'help': 'Where to generate the project'},
        {'name': 'preset_cookiecutter_json', 'flag': '-p', 'is_path': True, 'pre_exist': True, 'help': 'Cookiecutter JSON file path with all options provided in the file'},
        {'name': 'no_input', 'flag': '-i', 'action': 'store_true', 'help': 'Do not prompt for parameters and only use cookiecutter.json file content'},
        {'name': 'replay', 'flag': '-r', 'action': 'store_true', 'help': 'Do not prompt for parameters and only use information entered previously'},
        {'name': 'overwrite_if_exists', 'flag': '-f', 'action': 'store_true', 'help': 'Overwrite the contents of any preexisting project directory of the same name'},
        {'name': 'cookiecutter_version', 'flag': '-v', 'action': 'store_true', 'help': 'Cookiecutter version'},
    ]

    def run(self):

        if cookiecutter_ver is None:
            raise OasisException("Optional package 'cookiecutter' is not installed, to install oasislmf with extra packages run 'pip install oasislmf[extra]'")

        cmd_str = 'cookiecutter'
        if not self.cookiecutter_version:
            cmd_str += ''.join([
                (' --no-input' if self.no_input or self.preset_cookiecutter_json else ''),
                (' --replay' if self.replay else ''),
                (' --overwrite-if-exists' if self.overwrite_if_exists else ''),
                (' --output-dir {}'.format(self.output_dir) if self.output_dir else ''),
                (' --verbose ' if (self.logger.level < 20) else ' '),
                self.cookiecutter_uri()
            ])
        else:
            cmd_str += ' -V'

        self.logger.info('\nRunning cookiecutter command: {}\n'.format(cmd_str))

        try:
            run(cmd_str.split(), check=True)
        except CalledProcessError as e:
            self.logger.error(e)

    def cookiecutter_uri(self):
        raise NotImplementedError('Methode run must be implemented, this method returns the cookiecutter template url')


class CreateModelRepo(CreateModelBase):
    """
    Creates a local Git repository for a "simple model" (using the
    ``cookiecutter`` package) on a "simple model" repository template
    on GitHub)
    """
    def cookiecutter_uri(self):
        return 'git+ssh://git@github.com/OasisLMF/CookiecutterOasisSimpleModel'


class CreateComplexModelRepo(CreateModelBase):
    """
    Creates a local Git repository for a "complex model" (using the
    ``cookiecutter`` package) on a "simple model" repository template
    on GitHub)
    """
    def cookiecutter_uri(self):
        return 'git+ssh://git@github.com/OasisLMF/CookiecutterOasisComplexModel'
